﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace David.Academy.ProgrammingInCSharp.FunctionalProgramming
{
	/// <summary>
	/// Главен клас на програмата
	/// </summary>
	class Program
	{
		/// <summary>
		/// Главна програма
		/// </summary>
		/// <param name="args">Аргументи от командния ред</param>
		static void Main(string[] args)
		{
			// Въвеждане на цяло число от конзолата
			Console.Write("N = ");
			int n = int.Parse(Console.ReadLine());

			// Генериране на 10 случайни числа между 0 и 99
			int[] numbers = new int[10];
			Random random = new Random();
			for (int i = 0; i < 10; i++)
			{
				numbers[i] = random.Next(100);
			}

			// Изпечатване на самите числа
			PrintNumbers(numbers, x => x);
			// Изпечатване на числата, умножени по N
			PrintNumbers(numbers, x => x * n);
			// Изпечатване на числата, разделени на N
			PrintNumbers(numbers, x => x / n);
			// Изпечатване на числата, събрани с N
			PrintNumbers(numbers, x => x + n);
			// Изпечатване на числата, повдигнати на степен N
			PrintNumbers(numbers, x => (int)Math.Pow(x, n));
			// Изпечатване на квадратен корен от числата, закръглен надолу до цяло число
			PrintNumbers(numbers, x => (int)Math.Sqrt(x));
			// Изпечатване на сумата на числата
			Console.WriteLine(Sum(numbers));
			// Изпечатване на произведението на числата
			Console.WriteLine(Product(numbers));
			// Изпечатване на числата, разделени с интервали
			Console.WriteLine(Concat(numbers));

			// Пример за използване на вградените в .NET Framework методи от по-висок ред
			// за работа с обекти, които имплементират IEnumerable<T>;
			// 1. Умножава всяко число по 2;
			// 2. Отсява само резултатите, по-големи от 10
			// 3. Събира резултатите от стъпка 2 и ги присвоява на цяла променлива
			int sum = numbers
				.Select(x => x * 2)
				.Where(x => x > 10)
				.Aggregate(0, (a, i) => a + i);

			Console.ReadLine();
		}

		/// <summary>
		/// Изпечатва на конзолата масив от цели числа, като преди това ги преобразува с подаден
		/// като аргумент делегатен екземпляр и премахва нечетните резултати от преобразуването
		/// </summary>
		/// <param name="numbers">Масив от цели числа</param>
		/// <param name="transform">Делегатен екземпляр за преобразуване на числата</param>
		static void PrintNumbers(int[] numbers, Func<int, int> transform)
		{
			// Прилагане на подадения делегатен екземпляр върху всяко цяло число в масива
			IEnumerable<int> numbers2 = Select(numbers, transform);
			// Премахване на нечетните резултати от преобразуването
			IEnumerable<int> numbers3 = Where(
				numbers2,
				x => x % 2 == 0);

			// Изпечатване на получените резултати, разделени с интервал
			foreach (int number in numbers3)
			{
				Console.Write("{0} ", number);
			}
			// Преминаване на нов ред
			Console.WriteLine();
		}

		/// <summary>
		/// Прилага делегатен екземпляр върху всеки елемент в изброима колекция
		/// и връща изброима колекция от резултатите в същия ред;
		/// .NET Framework предоставя наготово аналогичен метод
		/// </summary>
		/// <typeparam name="T">Тип на елементите на входната колекция</typeparam>
		/// <typeparam name="TResult">Тип на елементите на резултатната колекция</typeparam>
		/// <param name="list">Входна изброима колекция от елементи</param>
		/// <param name="selector">Делегатен екземпляр за преобразуване на елементите</param>
		/// <returns>Резултатна изброима колекция</returns>
		static IEnumerable<TResult> Select<T, TResult>(
			IEnumerable<T> list, Func<T, TResult> selector)
		{
			List<TResult> results = new List<TResult>();
			foreach (T item in list)
			{
				results.Add(selector(item));
			}
			return results;
		}

		/// <summary>
		/// Проверява дали условие е изпълнено за всеки елемент от изброима колекция (поотделно)
		/// и връща изброима колекция от елементите, за които това е така;
		/// .NET Framework предоставя наготово аналогичен метод
		/// </summary>
		/// <typeparam name="T">Тип на елементите на входната и изходната колекции</typeparam>
		/// <param name="list">Входна изброима колекция</param>
		/// <param name="predicate">Делегатен екземпляр за условие, което елементът трябва да изпълнява</param>
		/// <returns>Резултатна изброима колекция</returns>
		static IEnumerable<T> Where<T>(
			IEnumerable<T> list, Func<T, bool> predicate)
		{
			List<T> results = new List<T>();
			foreach (T item in list)
			{
				if (predicate(item))
				{
					results.Add(item);
				}
			}
			return results;
		}

		/// <summary>
		/// Получава единствена агрегирана стойност от изброима колекция от елементи,
		/// прилагайки последователно делегатен екземпляр върху всеки от елементите ѝ
		/// и резултатната стойност до момента;
		/// .NET Framework предоставя наготово аналогичен метод
		/// </summary>
		/// <typeparam name="T">Тип на елементите на входната колекция</typeparam>
		/// <typeparam name="TResult">Тип на резултата</typeparam>
		/// <param name="list">Входна изброима колекция</param>
		/// <param name="aggregateFunction">Делегатен екземпляр за агрегиране на елементите;
		/// първият му аргумент е резултатната стойност до момента,
		/// а вторият - следващия елемент в изброимата колекция</param>
		/// <param name="startValue">Начална резултатна стойност</param>
		/// <returns>Резултатната стойност от агрегирането</returns>
		static TResult Aggregate<T, TResult>(
			IEnumerable<T> list,
			Func<TResult, T, TResult> aggregateFunction,
			TResult startValue)
		{
			TResult result = startValue;
			foreach (T item in list)
			{
				result = aggregateFunction(result, item);
			}
			return result;
		}

		/// <summary>
		/// Намира сумата на целите числа в изброима колекция
		/// </summary>
		/// <param name="numbers">Изброима колекция от цели числа</param>
		/// <returns>Сума на числата в колекцията</returns>
		static int Sum(IEnumerable<int> numbers)
		{
			// Прилагане на метода от по-висок ред
			// Aggregate<int, int>(IEnumerable<int>, Func<int, int, int>, int)
			// с агрегиращ делегатен екземпляр, който прибавя следващото число
			// към резултата до момента, и начална стойност, равна на 0
			return Aggregate<int, int>(
				numbers,
				(a, i) => a + i,
				0);
		}

		/// <summary>
		/// Намира произведението на целите числа в изброима колекция
		/// </summary>
		/// <param name="numbers">Изброима колекция от цели числа</param>
		/// <returns>Произведение на числата в колекцията</returns>
		static int Product(IEnumerable<int> numbers)
		{
			// Прилагане на метода от по-висок ред
			// Aggregate<int, int>(IEnumerable<int>, Func<int, int, int>, int)
			// с агрегиращ делегатен екземпляр, който умножава резултата до момента
			// по следващото число, и начална стойност, равна на 1
			return Aggregate<int, int>(
				numbers,
				(a, i) => a * i,
				1);
		}

		/// <summary>
		/// Конкатенира представянията в символни низове на целите числа в изброима
		/// колекция, разделяйки ги с интервали
		/// </summary>
		/// <param name="numbers">Изброима колекция от цели числа</param>
		/// <returns>Конкатенираният низ</returns>
		static string Concat(IEnumerable<int> numbers)
		{
			// Прилагане на метода от по-висок ред
			// Aggregate<int, string>(IEnumerable<int>, Func<string, int, string>, int)
			// с агрегиращ делегатен екземпляр, който конкатенира резултата до момента
			// с представянето на следващото число в символен низ, разделяйки ги с интервал,
			// и начална стойност, равна на празния низ
			return Aggregate<int, string>(
				numbers,
				(a, i) => a + " " + i.ToString(),
				string.Empty);
		}
	}
}
