﻿using System;
using System.Collections.Generic;

namespace David.Academy.ProgrammingInCSharp.OrcsAndDungeons
{
	/// <summary>
	/// Основен клас на играта
	/// </summary>
	internal class Game
	{
		private readonly int _width;
		private readonly int _height;
		private readonly Tile[,] _gameField;
		private readonly Random _random;
		private readonly Player _player;
		private readonly List<Monster> _monsters;

		private bool _isGameOver;

		/// <summary>
		/// Генератор на случайни числа
		/// </summary>
		public Random Random
		{
			get { return _random; }
		}

		/// <summary>
		/// Ширина на игралното поле
		/// </summary>
		public int Width
		{
			get { return _width; }
		}

		/// <summary>
		/// Височина на игралното поле
		/// </summary>
		public int Height
		{
			get { return _height; }
		}

		public Player Player
		{
			get { return _player; }
		}

		/// <summary>
		/// Създава екземпляр на играта
		/// </summary>
		/// <param name="width">Ширина на игралното поле</param>
		/// <param name="height">Височина на игралното поле</param>
		/// <param name="wallPercent">Процент стени в игралното поле</param>
		public Game(int width, int height, int wallPercent, int monsterCount)
		{
			_width = width;
			_height = height;
			_gameField = new Tile[width, height];
			_random = new Random();
			InitializeField(wallPercent);

			_player = new Player(this);
			_player.UnitDied += Player_UnitDied;

			_monsters = new List<Monster>();
			for (int i = 0; i < monsterCount; i++)
			{
				Monster monster = new Orc(this);
				monster.UnitDied += Monster_UnitDied;
				_monsters.Add(monster);
			}
		}

		void Monster_UnitDied(Unit unit)
		{
			Monster monster = (Monster)unit;
			PrintMessage(string.Format("{0} has died.", monster.Name));
			_monsters.Remove(monster);

			if (_monsters.Count == 0)
				_isGameOver = true;
		}

		void Player_UnitDied(Unit unit)
		{
			_isGameOver = true;
		}

		/// <summary>
		/// Основен цикъл на играта
		/// </summary>
		public void Run()
		{
			do
			{
				// Изпечатване на полето
				DrawField();
				PrintPlayerHitPoints();

				// Изчакване на действие от страна на играча
				ConsoleKeyInfo key = Console.ReadKey(true);
				ClearMessage();
				switch (key.Key)
				{
					case ConsoleKey.UpArrow:
						_player.MoveOrAttack(Direction.North);
						break;
					case ConsoleKey.DownArrow:
						_player.MoveOrAttack(Direction.South);
						break;
					case ConsoleKey.LeftArrow:
						_player.MoveOrAttack(Direction.West);
						break;
					case ConsoleKey.RightArrow:
						_player.MoveOrAttack(Direction.East);
						break;
					case ConsoleKey.Escape:
						_isGameOver = true;
						break;
				}

				if (!_isGameOver)
				{
					// Програмно управление на чудовищата на всеки ход
					foreach (Monster monster in _monsters)
					{
						monster.PerformAction();
					}
				}
			}
			while (!_isGameOver);

			if (_player.IsDead)
			{
				Console.Clear();
				Console.WriteLine("You died. Game over.");
				Console.ReadKey();
			}
			else if (_monsters.Count == 0)
			{
				Console.Clear();
				Console.WriteLine("You killed all the monsters. Congratulations!");
				Console.ReadKey();
			}
		}

		/// <summary>
		/// Взима единицата, която се намира на посочени координати
		/// </summary>
		/// <param name="point">Координати на точката</param>
		/// <returns>Единица, която се намира на указаното поле, иначе null</returns>
		public Unit GetTileUnit(Point point)
		{
			if (_player != null && point == _player.Point)
			{
				return _player;
			}

			if (_monsters != null)
			{
				foreach (Monster monster in _monsters)
				{
					if (point == monster.Point)
					{
						return monster;
					}
				}
			}

			return null;
		}

		/// <summary>
		/// Проверява дали плочка от игралното поле е празна
		/// </summary>
		/// <param name="x">X-координата на точката</param>
		/// <param name="y">Y-координата на точката</param>
		/// <returns></returns>
		public bool IsTileEmpty(Point point)
		{
			return (_gameField[point.X, point.Y] == Tile.Empty);
		}

		/// <summary>
		/// Генерира игралното поле
		/// </summary>
		/// <param name="wallPercent">Процент стени в игралното поле</param>
		private void InitializeField(int wallPercent)
		{
			for (int y = 0; y < _height; y++)
			{
				_gameField[0, y] = Tile.Wall;
				_gameField[_width - 1, y] = Tile.Wall;
			}
			for (int x = 0; x < _width; x++)
			{
				_gameField[x, 0] = Tile.Wall;
				_gameField[x, _height - 1] = Tile.Wall;
				for (int y = 1; y < _height - 1; y++)
				{
					if (_random.Next(100) < wallPercent)
						_gameField[x, y] = Tile.Wall;
				}
			}
		}

		/// <summary>
		/// Изчертава игралното поле на екрана
		/// </summary>
		private void DrawField()
		{
			// Подготваме буфер от символи, които да бъдат изпечатани
			char[,] chars = new char[_width, _height];
			for (int y = 0; y < _height; y++)
			{
				for (int x = 0; x < _width; x++)
				{
					switch (_gameField[x, y])
					{
						case Tile.Empty:
							chars[x, y] = ' ';
							break;
						case Tile.Wall:
							chars[x, y] = '\u2593';
							break;
					}
				}
			}
			chars[_player.Point.X, _player.Point.Y] = _player.Symbol;
			foreach (Monster monster in _monsters)
			{
				chars[monster.Point.X, monster.Point.Y] = monster.Symbol;
			}

			// Изпечатваме подготвените символи в конзолата
			Console.SetCursorPosition(0, 0);
			for (int y = 0; y < _height; y++)
			{
				for (int x = 0; x < _width; x++)
				{
					Console.Write(chars[x, y]);
				}
				Console.WriteLine();
			}
		}

		private void PrintPlayerHitPoints()
		{
			Console.SetCursorPosition(0, Height);
			Console.Write("Player hit points: {0,4}", _player.HitPoints);
		}

		private void PrintMessage(string message)
		{
			Console.SetCursorPosition(0, Height + 1);
			Console.Write(message);
		}

		private void ClearMessage()
		{
			Console.SetCursorPosition(0, Height + 1);
			for (int x = 0; x < Width - 1; x++)
			{
				Console.Write(' ');
			}
		}
	}
}
